@@ -1,8 +1,9 @@
 """ds4drv - A DualShock 4 bluetooth driver for Linux."""
 
 __title__ = "ds4drv"
-__version__ = "0.2.0"
+__version__ = "0.2.1"
 __author__ = "Christopher Rosell"
+__credits__ = ["Christopher Rosell", "George Gibbs", "Lauri Niskanen"]
 __license__ = "MIT"
 
 
@@ -34,10 +35,14 @@ L2CAP_PSM_HIDP_INTR = 0x13
 HIDP_TRANS_GET_REPORT = 0x40
 HIDP_TRANS_SET_REPORT = 0x50
 
-HIDP_DATA_RTYPE_INPUT = 0x01
-HIDP_DATA_RTYPE_OUTPUT = 0x02
+HIDP_DATA_RTYPE_INPUT   = 0x01
+HIDP_DATA_RTYPE_OUTPUT  = 0x02
 HIDP_DATA_RTYPE_FEATURE = 0x03
 
+BATTERY_MAX          = 8
+BATTERY_MAX_CHARGING = 11
+BATTERY_WARNING      = 2
+
 S16LE = Struct("<h")
 
 DS4Controller = namedtuple("DS4Controller", "id joystick options dynamic")
@@ -460,12 +465,43 @@ class DS4Device(object):
         self.ctl_sock = ctl_sock
         self.int_sock = int_sock
 
-        self.control(led_red=255, led_green=255, led_blue=255)
+        self._led = (0, 0, 0)
+        self._led_flash = (0, 0)
+        self._led_flashing = False
+
+        self.set_led(255, 255, 255)
+
+    def _control(self, **kwargs):
+        self.control(led_red=self._led[0], led_green=self._led[1],
+                     led_blue=self._led[2], flash_led1=self._led_flash[0],
+                     flash_led2=self._led_flash[1], **kwargs)
 
     def close(self):
         self.int_sock.close()
         self.ctl_sock.close()
 
+    def rumble(self, small=0, big=0):
+        self._control(small_rumble=small, big_rumble=big)
+
+    def set_led(self, red=0, green=0, blue=0):
+        self._led = (red, green, blue)
+        self._control()
+
+    def start_led_flash(self, on, off):
+        if not self._led_flashing:
+            self._led_flash = (on, off)
+            self._led_flashing = True
+            self._control()
+
+    def stop_led_flash(self):
+        if self._led_flashing:
+            self._led_flash = (0, 0)
+            self._led_flashing = False
+            # Call twice, once to stop flashing...
+            self._control()
+            # ...and once more to make sure LED is set.
+            self._control()
+
     def control(self, big_rumble=0, small_rumble=0,
                 led_red=0, led_green=0, led_blue=0,
                 flash_led1=0, flash_led2=0):
@@ -617,7 +653,114 @@ class ControllerAction(argparse.Action):
         namespace.controllers.append(controller)
 
 
+class ReportTimer(object):
+    def __init__(self, interval):
+        self.interval = interval
+        self.time = time()
+
+    def passed(self):
+        now = time()
+        if (now - self.time) > self.interval:
+            self.time = now
+            return True
+
+        return False
+
+
+class ReportAction(object):
+    def __init__(self, device, controller):
+        self.device = device
+        self.controller = controller
+        self.timers = {}
+
+    def add_timer(self, interval, func):
+        self.timers[func] = ReportTimer(interval)
+
+    def handle_report(self, report):
+        for func, timer in self.timers.items():
+            if timer.passed():
+                repeat = func(report)
+                if not repeat:
+                    self.timers.pop(func, None)
+
+
+class ReportActionBattery(ReportAction):
+    def __init__(self, device, controller):
+        super(ReportActionBattery, self).__init__(device, controller)
+
+        self.add_timer(60, self.check_battery)
+
+    def check_battery(self, report):
+        if report.battery < BATTERY_WARNING and not report.plug_usb:
+            self.device.start_led_flash(30, 30)
+            self.add_timer(5, lambda r: self.device.stop_led_flash())
+
+        return True
+
+
+class ReportActionJoystick(ReportAction):
+    def handle_report(self, report):
+        self.controller.joystick.emit(report)
+
+
+class ReportActionStatus(ReportAction):
+    def __init__(self, device, controller):
+        super(ReportActionStatus, self).__init__(device, controller)
+
+        self.report = None
+        self.add_timer(1, self.check_status)
+
+    def check_status(self, report):
+        if not self.report:
+            self.report = report
+            show_battery = True
+        else:
+            show_battery = False
+
+        # USB cable
+        if self.report.plug_usb != report.plug_usb:
+            plug_usb = report.plug_usb and "Connected" or "Disconnected"
+            show_battery = True
+
+            Daemon.info("USB: {0}", plug_usb,
+                        subprefix=CONTROLLER_LOG.format(self.controller.id))
+
+        # Battery level
+        if self.report.battery != report.battery or show_battery:
+            max_value = report.plug_usb and BATTERY_MAX_CHARGING or BATTERY_MAX
+            battery = 100 * report.battery // max_value
+
+            if battery < 100:
+                Daemon.info("Battery: {0}%", battery,
+                            subprefix=CONTROLLER_LOG.format(self.controller.id))
+            else:
+                Daemon.info("Battery: Fully charged",
+                            subprefix=CONTROLLER_LOG.format(self.controller.id))
+
+        # Audio cable
+        if (self.report.plug_audio != report.plug_audio or
+            self.report.plug_mic != report.plug_mic):
+
+            if report.plug_audio and report.plug_mic:
+                plug_audio = "Headset"
+            elif report.plug_audio:
+                plug_audio = "Headphones"
+            elif report.plug_mic:
+                plug_audio = "Mic"
+            else:
+                plug_audio = "Speaker"
+
+            Daemon.info("Audio: {0}", plug_audio,
+                        subprefix=CONTROLLER_LOG.format(self.controller.id))
+
+        self.report = report
+
+        return True
+
+
 def hexcolor(color):
+    color = color.strip("#")
+
     if len(color) != 6:
         raise ValueError
 
@@ -755,33 +898,17 @@ def find_devices():
 
 
 def read_device(device, controller):
-    options = controller.options
-    device.control(led_red=options.led[0],
-                   led_green=options.led[1],
-                   led_blue=options.led[2])
+    device.set_led(*controller.options.led)
+
+    actions = [cls(device, controller) for cls in
+               (ReportActionJoystick, ReportActionStatus)]
+
+    if controller.options.battery_flash:
+        actions.append(ReportActionBattery(device, controller))
 
-    led_last_flash = time()
-    led_flashing = True
     for report in device.reports:
-        if options.battery_flash:
-            if report.battery < 2 and not report.plug_usb:
-                if not led_flashing and (time() - led_last_flash) > 60:
-                    device.control(led_red=options.led[0],
-                                   led_green=options.led[1],
-                                   led_blue=options.led[2],
-                                   flash_led1=30, flash_led2=30)
-                    led_flashing = True
-                    led_last_flash = time()
-
-            if led_flashing and (time() - led_last_flash) > 5:
-                device.control(flash_led1=0, flash_led2=0)
-                device.control(led_red=options.led[0],
-                               led_green=options.led[1],
-                               led_blue=options.led[2])
-                led_flashing = False
-
-
-        controller.joystick.emit(report)
+        for action in actions:
+            action.handle_report(report)
 
     Daemon.info("Disconnected",
                 subprefix=CONTROLLER_LOG.format(controller.id))
@@ -828,3 +955,6 @@ def main():
         thread.controller = controller
         thread.start()
         threads.append(thread)
+
+if __name__ == "__main__":
+    main()
